/*
http://www.bbc.co.uk/blogs/researchanddevelopment/2010/08/new-developments-on-freeview-a.shtml
ETSI TS 102 323 V1.4.1 (2010-01)

Related content section, pid=0x????, table_id=0x76
Content identifier section, pid=0x0012, table_id=0x77

An RCT sub_table is carried in an elementary stream whose PID is identified by the presence of a related content
descriptor in the corresponding elementary stream descriptor loop of the current service's PMT (see clause 10.3). The
stream_type for an elementary stream carrying an RCT sub_table must be set to 0x05, indicating private sections (see
ISO/IEC 13818-1 [9], table 2-29).

This tap will:
    examine the pmt for the rct_pid.  this is carried as a private stream, component_tag=0x96 (always 96?)
    capture the rct section using the rct_pid, table_id=0x76
        these are normally empty, except during a trailer
    while active show the green button icon, and enable the green button
    on green button pressed,
        mach the series crid with the epg
        match the epsiode crid with the epg
        attempt to set a recording based on the epg entry found
        firmware should ask if a single/series record is wanted

- disabled excpet on live tv (firmware does not record rcts, fix this?...)
- not all trailers are flagged
- bbc1, bbc2, bbc3, bbc4 ok
- ch4, ch4+1, e4, more4 ok
- itv1,2,3,4, e4+1, ch5, others do not carry rct's

- only tested on 14.08
- should be ok on 14.09
- won't work on any other firmwares

*/
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"
#ifdef CallFirmwareA
#undef CallFirmwareA
#endif
#ifdef CallFirmwareA64
#undef CallFirmwareA64
#endif
#ifdef CallFirmwareT
#undef CallFirmwareT
#endif
#include "libFirebird.h"
#include "tsr_commander.h"


#define ID_USERTAP 0xbdb04301
TAP_ID( ID_USERTAP );

TAP_PROGRAM_NAME("Green Button");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("A trailer booking TAP v0.01");
TAP_ETCINFO(__DATE__);

#include "TSRCommander.inc"
//volatile register int *__gp asm("$28");

//----------------------------------------------------------------------------
type_menu main_menu;

int TAP_Main(void);
int close(void);
dword TAP_EventHandler(word event, dword param1, dword param2);
int main_menu_handler(type_menu *menu, dword param1);
int key_handler(dword param1, dword param2);
//----------------------------------------------------------------------------
void *fw_QueryRemote;
void *fw_DevDemux_Start;
void *fw_DBGSI_LoadDmx;
void *fw_free_filter;
void *fw_record_highlighted_event;
int set_fw_addresses();
//----------------------------------------------------------------------------
//greeen_button.ini settings
int ini_dirty = 0;
int splash_enable = 1;              //1=enable the initial splash osd
int menu_enable = 1;                //1=enable the menu
int menu_key = RKEY_Recall;         //1=enable the menu key
int debug = 0;                      //1=enable serial debug messages
dword fgcolour = COLOR_White;
dword bkcolour = RGB(3,5,10);

int ini_init();
int ini_update();
//----------------------------------------------------------------------------
int demux_hook_enabled = 0;
int demux_hook_id = -1;
int apply_demux_hook(void);
int remove_demux_hook(void);
int demux_hook(int param1, int param2);
//----------------------------------------------------------------------------
int five_second_enabled = 1;
dword five_second = 0;
void five_second_update(void);
//----------------------------------------------------------------------------
int green_button_enabled = 0;
dword pmt_handle = 0;
dword rct_handle = 0;
//----------------------------------------------------------------------------
#define DESCLEN 0x40
typedef struct {
    int   id;
    char  description[DESCLEN];
} EnumType;
//----------------------------------------------------------------------------
int process_PMT(byte *section);
int process_RCT(byte *section);
int set_section_filter(int ch, int pid, int tid, dword *handle);
int task_pmt(int handle, int value);
int task_rct(int handle, int value);
unsigned char *dvb_enumtype(EnumType *qenum, int id);
int to_name(char *dest_str, char *source_str, int len);
int do_desc(unsigned char *pkt, int stream_type, int link_id);
int get_pmt_info(unsigned char *buf);
int get_rct_info(unsigned char *buf);
//extern int dump_rct(FILE *Op, unsigned char *buf);

type_eventtable *find_crid_in_epg(int crid_type, char *crid, int *pevent_index);
type_eventtable *find_valid_crid_in_epg(int crid_type, char *crid, int *pevent_index);
int set_timer_by_crid(int crid_type, char *crid);
void dump_event(type_eventtable *et, int event_index, char *str);
int highlight_event(type_eventtable *et);

void green();
//----------------------------------------------------------------------------
typedef struct {
    int rct_pid;
    int year_offset;
    int num_links;
    struct {
        int term_id;
        int group_id;
        int precedence;
        char crid[0x20];
        char promotional_text[0x100];
        int default_icon_flag;
        int icon_id;
        char event_name[0x40];
    } link[4];
} type_related_content;
type_related_content related_content;

//----------------------------------------------------------------------------
#define b150(buf,j)  (((buf[j] << 8) & 0xff00) + (buf[j+1] & 0x00ff))
#define b120(buf,j)  (((buf[j] << 8) & 0x1f00) + (buf[j+1] & 0x00ff))
#define b110(buf,j)  (((buf[j] << 8) & 0x0f00) + (buf[j+1] & 0x00ff))
#define b94(buf,j)   (((buf[j] << 4) & 0x30) + ((buf[j+1] >> 4) & 0x0f))
#define b70(buf,j)   ((buf[j] >> 0) & 0xff)
#define b50(buf,j)   ((buf[j] >> 0) & 0x3f)
#define b30(buf,j)   ((buf[j] >> 0) & 0x0f)
#define b74(buf,j)   ((buf[j] >> 4) & 0x0f)
#define b64(buf,j)   ((buf[j] >> 4) & 0x07)
#define b7(buf,j)    ((buf[j] >> 7) & 0x01)

#define SHORT_EVENT_DESCRIPTOR                  0x004d
#define STREAM_IDENTIFIER_DESCRIPTOR            0x0052

//----------------------------------------------------------------------------
dword ShowMessage_xy2(char *lpMessage, int x, int y, dword fgcolour, dword bkcolour, dword dwDelay)
{
    word rgn;
    int w;

    //delete any old ones
//but not if a menu is showing?

    //make sure we have a region
    if (message_rgn == 0) {
        rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
        if ((short)rgn == -1) {
//            printf("\nShowMessage_xy(rgn=%x)", rgn);
            return 0;
        }
        message_rgn = rgn;
    }
    else {
        rgn = message_rgn;
    }

    //print the message
    w = TAP_Osd_GetW(lpMessage, 0, FNT_Size_1622) + 10; // calculate width of message
    w = ((x+w)>715) ? 715-x : w;
    TAP_Osd_PutString(rgn, x, 5+y+22, x+w, lpMessage, fgcolour, bkcolour, 0, FNT_Size_1622, 1);

    //schedule delete
    message_showing = TAP_GetTick() + dwDelay;
    return message_rgn;
}

//----------------------------------------------------------------------------
#define P32 0xedb88320  //crc polynomial
dword fcstab8[0x100];

dword bitrevN(int n, dword data)
{
    int i;
    dword result;

    result = 0;
    for (i=0; i<n; i++) {
        result += ((data >> (n-1-i)) & 1) << (i);
    }
    return result;
}

void genfcs(int dsize)
{
    int b;
    int b2;
    dword v;
    int i;

    for (b=0; b<(1<<dsize); b++) {
        b2 = bitrevN(8,b);
        v = (dword)b;
        for(i=0; i<dsize; i++) {
            v = (v >> 1) ^ ((v & 1) ? P32 : 0);
        }

        fcstab8[b2] = bitrevN(32,v);
    }
}

// 32 bit crc, 8 bit data, 8 bits at a time
// lookup table = 2^8=256 words
dword crc8(dword crc, byte *data, int len)
{
    int i;
    int index;
    dword dataout32;

    for(i=0; i<len; i++) {
        index = ((crc>>24) ^ (data[i])) & 0xff;
        dataout32 = fcstab8[index];
        crc = (crc << 8) ^ dataout32;
    }

    return crc;
}

//----------------------------------------------------------------------------
char *tap_name;
char menu_title_str[0x40];

type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {5, "xxx",          },
        {0, "Status",       },
        {1, "Disable Green Button", },
        {2, "Enable Serial Debug",  },
        {3, "Disable Menu",         },
        {4, "Exit Tap",    },
    },
};

//----------------------------------------------------------------------------

int TAP_Main(void)
{
    int ret;
    char filename[] = "green_button.log";

    tap_name =  (char *)__tap_program_name__;
    if (debug) sprintf(strLog, "Opening %s", tap_name);
    if (debug) log_message(LOG_SERIAL, strLog, DELAY_SHORT);

    ini_init();

    if (splash_enable) sprintf(strLog, "Opening %s", tap_name);
    if (splash_enable) log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);

    ret = exTAP_Init();
    if (ret < 0) {
        sprintf(strLog, "exTAP_INIT: Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return 0;
    }
    ret = set_fw_addresses();
    if (ret < 0) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return 0;
    }

    genfcs(8);
    main_menu.visible = 0;
    sprintf(menu_title_str, "%s", tap_name);
    main_menu.fixme_titlep = menu_title_str - _tap_startAddr;
    strcpy(main_menu.item[3].str, (debug) ? "Disable Serial Debug" : "Enable Serial Debug");
    strcpy(main_menu.item[4].str, (menu_enable) ? "Disable Menu" : "Enable Menu");

    if (!OpLog) {
        TAP_Hdd_ChangeDir("/");
        TAP_Hdd_ChangeDir("ProgramFiles");
        //if exists delete it
        if (TAP_Hdd_Exist(filename)) {
            TAP_Hdd_Delete(filename);
        }
        TAP_Hdd_Create(filename, ATTR_NORMAL);
        OpLog = TAP_Hdd_Fopen(filename);
        if (!OpLog) {
            if (debug & 1) sprintf(strLog, "ERROR: cannot open logfile, %s", filename);
            if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        }
    }

    TSRCommanderInit(0, TRUE);
    five_second = TAP_GetTick()+ (5*100);
    return 1;                 //0=application mode, 1=tsr mode
}

int close(void)
{
    *(dword *)(EMMA_TTSD_SR_CMD) = 0;           //clear promiscuous
    five_second_enabled = 0;
    remove_demux_hook();

    DeleteMenu(&main_menu);

    if (ini_dirty) ini_update();

    if (splash_enable) sprintf(strLog, "Closing  %s", tap_name);
    if (splash_enable) log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

    if (message_rgn) TAP_Osd_Delete(message_rgn);

    if (OpLog) TAP_Hdd_Fclose(OpLog);

    TSRCommanderErase();
    if (debug) printf("\nClosing %s", tap_name);
    return 0;
}

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    dword tick;
    dword state, subState;

    switch (event) {
        case EVT_IDLE:
//            printf("\nEVT_IDLE, event=%x, param1=%x, param2=%x", event, param1, param2);
            tick = TAP_GetTick();

            //
            //message window
            //
            if (message_showing) {
                if (tick >= message_showing) {
                    TAP_EnterNormal();
                	TAP_Osd_Delete(message_rgn);
                    message_showing = 0;
                    message_rgn = 0;
                }
            }
            //
            //housekeeping
            //
            if (five_second_enabled) {
                if (tick > five_second) {
                    five_second = tick + (5*100);
                    five_second_update();
                }
            }

    		TSRCommanderWork();
            break;


        case EVT_KEY:
            TAP_GetState(&state, &subState);
            //printf("\nEVT_KEY, param1=%x, param2=%x, state=%s, subState=%s", param1, param2, get_str(state_str, state), get_str(substate_str, subState));

            if (main_menu.visible) {
                param1 = main_menu_handler(&main_menu, param1);
            }
            else
            if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                param1 = key_handler(param1, param2);
            }
	        break;

        case EVT_HOOK:
            if (param1 == demux_hook_id) param1 = demux_hook(param1, param2);
            break;

        case EVT_UART:
            break;

        default:
//            printf("\nEVT_???, event=%x, param1=%x, param2=%x", event, param1, param2);
            break;
    }
    return param1;
}

int main_menu_handler(type_menu *menu, dword param1)
{
    int selection;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:  //status
                    if (related_content.rct_pid == 0xffff) sprintf(strLog, "Related Content not present");
                    else if (related_content.rct_pid == 0x0000) sprintf(strLog, "PMT PID not analysed");
                    else sprintf(strLog, "Related Content present, rct_pid=%04x", related_content.rct_pid);
                    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
				    return 0;
				case 1: // enable/disable
                    five_second_enabled = (five_second_enabled) ? 0 : 1;
                    if (!five_second_enabled) {
                        remove_demux_hook();
                    }
                    strcpy(main_menu.item[2].str, (five_second_enabled) ? "Disable Green Button" : "Enable Green Button");
				    return 0;
				case 2: // enable/disable debug
                    debug = (debug) ? 0 : -1;
                    ini_dirty = 1;
                    strcpy(main_menu.item[3].str, (debug) ? "Disable Serial Debug" : "Enable Serial Debug");
				    return 0;
				case 3: //enable/disable menu
                    menu_enable = (menu_enable) ? 0 : 1;
                    ini_dirty = 1;
                    strcpy(main_menu.item[4].str, (menu_enable) ? "Disable Menu" : "Enable Menu");
				    return 0;
				case 4:
				    close();
				    TAP_Exit();
				    return 0;
			}
            break;
        case RKEY_Exit:
			DeleteMenu(menu);
            return 0;
        default:
            if (param1 == menu_key) {
                 DeleteMenu(menu);
                 return param1;
            }
	        TAP_Win_Action(&menu->win, param1);
    		return 0;
    }
    return param1;
}


int key_handler(dword param1, dword param2)
{
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];

    if ((param1 == menu_key) && (menu_enable)) {
        //only if tsrcommander is not running
        if (!is_tsr_commander_running()) {
            ShowMenu(&main_menu);
            return 0;
        }
    }
    switch(param1) {
        case RKEY_Green:
            if (green_button_enabled) {
                //only when live
                TAP_Hdd_GetPlayInfo(&pi);
                if (pi.playMode == PLAYMODE_None) {
                    green();
                }
                return 0;
            }
            break;
    }
    return param1;
}

//----------------------------------------------------------------------------
int set_fw_addresses()
{
    if (fw_sysid == 456) {
        switch(_appl_version) {
            case 0x1408:
                fw_QueryRemote = (void *)0x8000649c;
                fw_DevDemux_Start = (void *)0x8007df14;
                fw_DBGSI_LoadDmx = (void *)0x8007eb80;
                fw_free_filter = (void *)0x8007ed54;
                fw_record_highlighted_event = (void *)0x80126610;
                break;
            case 0x1409:
                fw_QueryRemote = (void *)0x8000649c;
                fw_DevDemux_Start = (void *)0x8007df14;
                fw_DBGSI_LoadDmx = (void *)0x8007eb80;
                fw_free_filter = (void *)0x8007ed54;
                fw_record_highlighted_event = (void *)0x801266B8;
                break;
            default:
                return -1;
        }

        //these are in exTap
        if (((dword)fw_event_table & 0xc0000000) != 0x80000000) return -1;
        if (fw_event_table_size <= 0) return -1;
        if (((dword)fw_epg_highlighted_event & 0xc0000000) != 0x80000000) return -1;
        return 0;
    }
    return -1;
}

//-----------------------------------------------------------------------------
int ini_init()
{
    int ini_location;

    ini_location = INIOpenFile("green_button.ini", "green_button");
    splash_enable = INIGetInt("splash_enable", splash_enable, 0x80000000, 0x7fffffff);
    menu_enable = INIGetInt("menu_enable", menu_enable, 0x80000000, 0x7fffffff);
    menu_key = INIGetHexDWord("menu_key", menu_key, 0x00000000, 0xffffffff);
    debug = INIGetHexDWord("debug", debug, 0x00000000, 0xffffffff);
    fgcolour = INIGetHexDWord("fgcolour", fgcolour, 0x00000000, 0xffffffff);
    bkcolour = INIGetHexDWord("bkcolour", bkcolour, 0x00000000, 0xffffffff);

    INICloseFile();
    ini_update();
    return 0;
}

int ini_update()
{
    int ini_location;

    ini_location = INIOpenFile("green_button.ini", "green_button");
    INISetInt("splash_enable", splash_enable);
    INISetInt("menu_enable", menu_enable);
    INISetHexDWord("menu_key", menu_key);
    INISetHexDWord("debug", debug);
    INISetHexDWord("fgcolour", fgcolour);
    INISetHexDWord("bkcolour", bkcolour);

    INISaveFile("green_button.ini", INILOCATION_AtSettings, "green_button");
    INICloseFile();
    ini_dirty = 0;
    return 0;
}

//----------------------------------------------------------------------------
//hook the end of fw_QueryRemote
//
type_hook demux_hook_patch[] = {
// sysid, firmware, enable, addr,       length, original_data
    {456, 0x1408,   1,      0x80006544, 8,      {0x32, 0xb5, 0xff, 0xff,  0x02, 0xa0, 0x10, 0x25, }, },  //end of fw_QueryRemote
};

int apply_demux_hook(void)
{
    //fix the address;
    demux_hook_patch[0].sysid = fw_sysid;
    demux_hook_patch[0].revision = _appl_version;
    demux_hook_patch[0].enable = 1;
    demux_hook_patch[0].addr = 0xa8+(dword)fw_QueryRemote;

    demux_hook_id = exTAP_ApplyHook(&demux_hook_patch[0]);
    if (demux_hook_id < 0) {
        demux_hook_enabled = 0;
        sprintf(strLog, "ERROR: apply_demux_hook(), firmware version %04x not supported", _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return -1;
    }

    demux_hook_enabled = 1;
    return 0;
}

int remove_demux_hook(void)
{
    int ret;

    if (demux_hook_id >=0) {
        //we must free up any active filters, otherwise we cannot reuse the pid
        if (pmt_handle) {
            if (debug & 4) sprintf(strLog,"fw_free_filter(pmt_handle = %x)", pmt_handle);
            if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            exTAP_CallFirmware((void *)fw_free_filter, pmt_handle, 0,0,0,0,0,0,0,0,0,0);
            pmt_handle = 0;
        }
        if (rct_handle) {
            if (debug & 4) sprintf(strLog,"fw_free_filter(rct_handle = %x)", rct_handle);
            if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            exTAP_CallFirmware((void *)fw_free_filter, rct_handle, 0,0,0,0,0,0,0,0,0,0);
            rct_handle = 0;
        }

        ret = exTAP_RemoveHook(demux_hook_id);
        if (ret == 0) demux_hook_id = -1;
        if (debug & 8) sprintf(strLog, "Demux Hook Removed");
        if (debug & 8) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    }

    demux_hook_enabled = 0;
    return 0;
}

//s5 contains the event handle
//s8/fp contains the address of the param
int demux_hook(int param1, int param2)
{
    type_hook_registers *reg;

//printf("\nEVT_HOOK_demux, param1=%x, param2=%x  ", param1, param2);
    reg = (type_hook_registers *)param2;
    if (reg->s5) {
        if (reg->s5 == pmt_handle) {
            task_pmt(reg->s5, *(dword *)reg->s8);
            //ideally, we should bin this now
            reg->s5 = 0;
        }
        else
        if (reg->s5 == rct_handle) {
            task_rct(reg->s5, *(dword *)reg->s8);
            //ideally, we should bin this now
            reg->s5 = 0;
        }
    }
    return param1;
}

//-----------------------------------------------------------------------------
int set_section_filter(int ch, int pid, int tid, dword *handle)
{
    unsigned long long tid64;
    unsigned long long mask64;
    unsigned long long tid_mask64;
    int hDemux;

    //make sure the filter is free
    if (*handle) {
        if (debug & 4) sprintf(strLog,"fw_free_filter(*handle = %x)", *handle);
        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        exTAP_CallFirmware((void *)fw_free_filter, *handle, 0,0,0,0,0,0,0,0,0,0);
        *handle = 0;
    }

    tid64 = (unsigned long long)tid;
    mask64 = 0xff;
    tid_mask64 = 0x00;

//DevDemux_Start(0, UCHAR ch, USHORT pid, 1, UINT64 tid, UINT64 mask, UINT64 tid_mask, ULONG timeout, ULONG nSec);
//               0,1   0,1,2              0,1,81                                       1/10 s         0,1,fe,ff

//bbc has ~1s  between empty rcts
//ch4 has ~30s between empty rcts
    //set a section filter
    if (debug & 4) sprintf(strLog,"fw_DevDemux_Start(0, %d, 0x%04x, 1, 0x%02x, 0x%02x, 0x%02x, 400, 0)", ch, pid, ((byte *)&tid64)[7], ((byte *)&mask64)[7], ((byte *)&tid_mask64)[7]);
    if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    hDemux = exTAP_CallFirmware((void *)fw_DevDemux_Start, 0, ch, pid, 1, (dword)&tid64, (dword)&tid_mask64, (dword)&mask64, 400, 0,  0,0);
    hDemux &= 0xffff;

    if (hDemux >= 0xe000) {
        *handle = 0;
        if (debug & 1) sprintf(strLog, "ERROR: fw_DevDemux_Start(), ret=%x", hDemux);
        if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        return -1;
    }
    else {
        if (debug & 1) printf(", ret=%x", hDemux);
    }
    *handle = hDemux;
    return 0;
}

//-----------------------------------------------------------------------------
int task_pmt(int handle, int value)
{
    int ret;
    byte nSec, nLoadedSec;
    byte **dscr;
    int i;
    int main_channel;

//    printf("\ntask_pmt(handle=%x, value=%x)", handle, value);
    //load the section + process
    if (value == 0) {
        if (debug & 4) sprintf(strLog,"fw_DBGSI_LoadDmx(%x)", handle);
        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        ret = exTAP_CallFirmware((void *)fw_DBGSI_LoadDmx, handle, (dword)&nSec, (dword)&nLoadedSec, (dword)&dscr,  0,0,0,0,0,0,0);
        ret &= 0xffff;
//        if (debug & 4) sprintf(strLog,"ret=%x", ret);
//        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);

        if (ret) {
            if (debug & 1) sprintf(strLog, "ERROR: fw_DBGSI_LoadDmx(), ret=%x", ret);
            if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            goto exit;
        }
        else {
            if (debug & 1) printf(", nSec = %x, nLoadedSec = %x, dscr = %x", nSec, nLoadedSec, dscr);
        }
        if (!is_firmware(dscr)) {
            if (debug & 1) sprintf(strLog, "ERROR: fw_DBGSI_LoadDmx(), ret=%x, dscr = %x", ret, dscr);
            if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            goto exit;
        }
//        dump(dscr, 0x40);
///what if we have multiple sections, nLoadedSec > 1 ?
        for (i=0; i<nLoadedSec; i++) {
            if (!is_firmware(dscr[i])) {
                if (debug & 1) sprintf(strLog, "ERROR: fw_DBGSI_LoadDmx(), ret=%x, dscr[%x] = %x", ret, i, dscr[i]);
                if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
                goto exit;
            }

//        dump(dscr[i], 0x40);
            ret = process_PMT(dscr[i]);
            if (ret) {
                goto exit;
            }
            if (debug & 8) sprintf(strLog,"rct_pid = %04x", related_content.rct_pid);
            if (debug & 8) log_message(LOG_SERIAL, strLog, DELAY_SHORT);

            //operate on the main channel
            main_channel = ((*(volatile dword *)EMMA_TTSD_1ST_VIDEO_PID) >> 29)&3;

            //now trigger the rct
            if (related_content.rct_pid && (related_content.rct_pid != 0xffff)) {
                ret = set_section_filter(main_channel, related_content.rct_pid, 0x76, &rct_handle);
            }
        }
    }
    else {
        //2 = timeout
        if (debug & 4) sprintf(strLog,"task_pmt(), handle=%x, value=%x", handle, value);
        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    }

exit:
    //free the filter
    if (pmt_handle) {
        if (debug & 4) sprintf(strLog,"fw_free_filter(pmt_handle=%x)", pmt_handle);
        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        exTAP_CallFirmware((void *)fw_free_filter, pmt_handle, 0,0,0,0,0,0,0,0,0,0);
        pmt_handle = 0;
    }
    return 0;
}
//-----------------------------------------------------------------------------

int task_rct(int handle, int value)
{
    int ret;
    byte nSec, nLoadedSec;
    byte **dscr;
    int i;
    int main_channel;

//    printf("\ntask_rct(handle=%x, value=%x)", handle, value);
    //load the section + process
    if (value == 0) {
        if (debug & 4) sprintf(strLog, "fw_DBGSI_LoadDmx(%x)", handle);
        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        ret = exTAP_CallFirmware((void *)fw_DBGSI_LoadDmx, handle, (dword)&nSec, (dword)&nLoadedSec, (dword)&dscr,  0,0,0,0,0,0,0);
        ret &= 0xffff;
//        if (debug & 4) sprintf(strLog,"ret=%x", ret);
//        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);

        if (ret) {
            if (debug & 1) sprintf(strLog, "ERROR: fw_DBGSI_LoadDmx(), ret=%x", ret);
            if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            goto exit;
        }
        else {
            if (debug & 1) printf(", nSec = %x, nLoadedSec = %x, dscr = %x", nSec, nLoadedSec, dscr);
        }
        if (!is_firmware(dscr)) {
            if (debug & 1) sprintf(strLog, "ERROR: fw_DBGSI_LoadDmx(), ret=%x, dscr = %x", ret, dscr);
            if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            goto exit;
        }
//        dump((dword *)dscr, 0x40);
///what if we have multiple sections, nLoadedSec > 1 ?
        for (i=0; i<nLoadedSec; i++) {
            if (!is_firmware(dscr[i])) {
                if (debug & 1) sprintf(strLog, "ERROR: fw_DBGSI_LoadDmx(), ret=%x, dscr[%x] = %x", ret, i, dscr[i]);
                if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
                goto exit;
            }

//        dump(dscr[i], 0x40);
            ret = process_RCT(dscr[i]);
            if (ret) {
                goto exit;
            }
        }
    }
    else {
        //2 = timeout
        if (debug & 4) sprintf(strLog,"task_rct(), handle=%x, value=%x", handle, value);
        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        if (value != 2) goto exit;
    }

    //free the filter
    if (rct_handle) {
        if (debug & 4) sprintf(strLog,"fw_free_filter(rct_handle = %x)", rct_handle);
        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        exTAP_CallFirmware((void *)fw_free_filter, rct_handle, 0,0,0,0,0,0,0,0,0,0);
        rct_handle = 0;
    }

    //operate on the main channel
    main_channel = ((*(volatile dword *)EMMA_TTSD_1ST_VIDEO_PID) >> 29)&3;

    //now trigger another rct
    if (related_content.rct_pid && (related_content.rct_pid != 0xffff)) {
        ret = set_section_filter(main_channel, related_content.rct_pid, 0x76, &rct_handle);
    }

    return 0;

exit:
    //free the filter
    if (rct_handle) {
        if (debug & 4) sprintf(strLog,"fw_free_filter(rct_handle = %x)", rct_handle);
        if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        exTAP_CallFirmware((void *)fw_free_filter, rct_handle, 0,0,0,0,0,0,0,0,0,0);
        rct_handle = 0;
    }
    return 0;
}
//-----------------------------------------------------------------------------

int process_PMT(byte *section)
{
    int section_length;
    dword crc;

//    if (debug & 4) sprintf(strLog,"process_PMT(%x)", section);
//    if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    //extract rct_pid
    section_length = 3+((*(dword *)section >> 8)&0xfff);
//    dump(section, section_length);

    //do crc - only for sections that have a crc ...
    crc = 0xffffffff;
    crc = crc8(crc, section, section_length);
    if (crc) {
        if (debug & 1) sprintf(strLog, "ERROR: process_PMT(), crc=%x", crc);
        if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        return -1;
    }

    related_content.rct_pid = 0xffff;
    get_pmt_info(section);
//    if (debug & 8) dump_pmt(NULL, section, 0);

    return 0;
}

int process_RCT(byte *section)
{
    int section_length;
    dword crc;
    int num_links;
    char str[0x40];
    int i;

//    if (debug & 4) sprintf(strLog,"process_RCT(%x)", section);
//    if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    //extract term_id, crid
    section_length = 3+((*(dword *)section >> 8)&0xfff);
//    dump(section, section_length);

    //do crc - only for sections that have a crc ...
    crc = 0xffffffff;
    crc = crc8(crc, section, section_length);
    if (crc) {
        if (debug & 1) sprintf(strLog, "ERROR: process_RCT(), crc=%x", crc);
        if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        return -1;
    }

//    if (debug & 8) printf("\n");
//    if (debug & 8) dump_rct(NULL, section);
    get_rct_info(section);

    if (debug & 0x8) {
        if (related_content.num_links) {
            sprintf(strLog,"rct_pid = %04x, num_links = %x", related_content.rct_pid, related_content.num_links);
            log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        }
        for (i=0; i<related_content.num_links; i++) {
            sprintf(strLog,"(%x)%s", related_content.link[i].term_id, related_content.link[i].crid);
            log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            sprintf(strLog,"%s. %s", related_content.link[i].event_name, related_content.link[i].promotional_text);
            log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        }
    }
    if (debug & 0x10) {
        if (related_content.num_links) {
            sprintf(strLog,"rct_pid = %04x, num_links = %x", related_content.rct_pid, related_content.num_links);
            log_message(LOG_FILE, strLog, DELAY_SHORT);
        }
        for (i=0; i<related_content.num_links; i++) {
            sprintf(strLog,"(%x)%s", related_content.link[i].term_id, related_content.link[i].crid);
            log_message(LOG_FILE, strLog, DELAY_SHORT);
            sprintf(strLog,"%s. %s", related_content.link[i].event_name, related_content.link[i].promotional_text);
            log_message(LOG_FILE, strLog, DELAY_SHORT);
        }
    }


    //if the rct is valid, return the crid(s)
    num_links = related_content.num_links;
    if (num_links > 0) {
        //dont want to reset the timeout on any existing messages
        if (!message_showing) {
            //enable GreenButton
            green_button_enabled = 1;
            //show GreenButton
            ShowMessage_xy2(OSD_GREEN, 40, 15, COLOR_White, COLOR_None, 3000);
            sprintf(str, " %s", related_content.link[0].event_name);
            ShowMessage_xy2(str, 75, 15, COLOR_White, COLOR_DarkGray, 3000);
        }
    }
    else {
        if (green_button_enabled) {
            //clear "Record Green"
            ShowMessage_xy2(OSD_GREEN, 40, 15, COLOR_White, COLOR_None, 10);
        }
        green_button_enabled = 0;
    }
    return 0;
}

//-----------------------------------------------------------------------------
unsigned char *dvb_enumtype(EnumType *qenum, int id)
{
    int i;

    for(i=1; i<=qenum[0].id; i++) {
        if (qenum[i].id == id) {
            return qenum[i].description;
        }
    }
    return qenum[0].description;
}

//ascii to utf
//  [0x00-0x7F]                  # Single-byte chars (ASCII range)
//  [0xC0-0xDF] [0x80-0xBF]      # Two-byte chars (0x0080-0x07FF)
//  [0xE0-0xEF] [0x80-0xBF]      # Three-byte chars (0x800-0xFFFF)
//  [0xF0-0xF4] [0x80-0xBF]      # Four-byte chars (0x10000-0x10FFFF)
//
int to_name(char *dest_str, char *source_str, int len)
{
    int ret;
    int i,j;

    if (len == 0) {
        dest_str[0] = '\0';
        return 0;
    }
    if (len == 1) {
        dest_str[0] = to_printable((unsigned char)source_str[0]);
        dest_str[1] = '\0';
    }
    ret = 0;
#ifdef FREESAT
    //huffman code = 0x1f 0x01
    //            or 0x1f 0x02
    if (source_str[0] == 0x1f) {
        ret = freesat_huffman_to_string((unsigned char *)source_str, dest_str, len);
    }
#endif
    if (ret==0) {
        j = 0;
        for (i=0; i<len; i++) {
            //ascii to utf
            if ((unsigned char)source_str[i] < 0x20) {                          //1 byte codes
                dest_str[j++] = 0x20;;
            }
            else
            if ((unsigned char)source_str[i] < 0x80) {                          //1 byte codes
                dest_str[j++] = source_str[i];
            }
            else {                                                              //2 byte codes
                dest_str[j++] = 0xc0 | ((source_str[i]&0xc0)>>6);
                dest_str[j++] = 0x80 | (source_str[i]&0x3f);
            }
        }
        dest_str[j++] = '\0';
    }
    return 0;
}

int do_desc(unsigned char *pkt, int stream_type, int link_id)
{
    int descriptor_tag, descriptor_length;
    int j;
    int component_tag = 0;
    char name[0x120];

    j = 0;
    descriptor_tag = b70(pkt, j++);
    descriptor_length = b70(pkt, j++);
    if (!descriptor_length) return 2;

    switch (descriptor_tag)  {
        case SHORT_EVENT_DESCRIPTOR:                    // 0x4d
            //only do the rct, not the eit
            if (stream_type == 0x05) {
//            if (component_tag == 0x96) {
                j+=3;
                to_name(name, (char *)&pkt[j+1], pkt[j+0]);
                j+=pkt[j+0];
                j++;
                strncpy(related_content.link[link_id].event_name, name, 0x40);
            }
            break;

        case STREAM_IDENTIFIER_DESCRIPTOR:              // 0x52
            component_tag = pkt[j++];
            switch(stream_type) {
            case 0x05:  //private section
                //how to pick rct pids ? - component_tag = 96
                if (component_tag == 0x96) {
                    related_content.rct_pid = link_id;
                }
                break;
            default:
                break;
            }
            break;

        default:
            break;
    }
    return 2+descriptor_length;
}

int get_pmt_info(unsigned char *buf)
{
    int service_id, elementary_pid, stream_type, pcr_pid;
    int section_length, descriptor_length;
    int j,k;

    j = 0;
    j++;
    section_length = 3 + b110(buf, j); j+=2;
    service_id = b150(buf, j); j+=2;
    j += 3;
    pcr_pid = b120(buf, j); j+=2;
    descriptor_length = b110(buf, j); j+=2;
    k = j;
    while (j < (k+descriptor_length)) {
        j += do_desc(&buf[j],-1,-1);
    }

    while (j<(section_length-4)) {
        stream_type           = b70(buf, j++);
        elementary_pid        = b120(buf, j); j+=2;
        descriptor_length     = b110(buf, j); j+=2;
        k = j;
        while (j < (k+descriptor_length)) {
            j += do_desc(&buf[j],stream_type,elementary_pid);
        }
    }

    return 0;
}

int get_rct_info(unsigned char *buf)
{
    int service_id;
    int section_length, descriptor_length;
    int i,j,k,l;
    int year_offset, link_count;
    int link_info_length;
    int link_type, how_related_classification_scheme_id, term_id, group_id, precedence;
    int media_uri_length, number_items, promotional_text_length;
    int default_icon_flag, icon_id;

    j = 0;
    j++;
    section_length = 3 + b110(buf, j); j+=2;
    service_id = b150(buf,j); j+=2;
    j+=3;

    year_offset = b150(buf,j); j+=2;
    related_content.year_offset = year_offset;
    link_count = buf[j++];
    related_content.num_links = link_count;
    if (link_count > 4) {
        if (debug & 1) sprintf(strLog, "ERROR: get_rct_info(), link_count=%x", link_count);
        if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        link_count = 0;
        return -1;
    }
    for (i=0; i<link_count; i++) {
        link_info_length = b110(buf,j); j+=2;
        link_type = b74(buf,j);
        how_related_classification_scheme_id = b94(buf,j);
        term_id = b110(buf,j+1); j+=3;                     //2=GroupTrailer, 5=Parent //ETSI TS 102 822-3-1 V1.5.1 (2009-05) ts_1028220301v010501p.pdf
        related_content.link[i].term_id = term_id;
        group_id = b74(buf,j);
        related_content.link[i].group_id = group_id;
        precedence = b30(buf,j); j++;
        related_content.link[i].precedence = precedence;

        if ((link_type == 0x00) || (link_type == 0x02)) {
            media_uri_length = buf[j++];
            for (k=0; k<media_uri_length; k++) {
                related_content.link[i].crid[k] = buf[j++];
            }
            related_content.link[i].crid[k] = '\0';
        }
        if ((link_type == 0x01) || (link_type == 0x02)) {
            if (debug & 1) sprintf(strLog, "ERROR: get_rct_info(), unhandled dvb_binary_locator()");
            if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            link_count = 0;
            return -1;
        }
        number_items  = b50(buf,j++);          //!!!what if more than 1 item?
        for (k=0; k<number_items; k++) {
            j+=3;
            promotional_text_length = buf[j++];
            for (l=0; l<promotional_text_length; l++) {
                related_content.link[i].promotional_text[l] = buf[j++];
            }
            related_content.link[i].promotional_text[l] = '\0';
        }
        default_icon_flag = b7(buf,j);
        related_content.link[i].default_icon_flag = default_icon_flag;
        icon_id = b64(buf,j);
        related_content.link[i].icon_id = icon_id;
        descriptor_length   = b110(buf,j); j+=2;
        k = j+descriptor_length;
        while (j<k) {
            j += do_desc(&buf[j],5,i);     //stream_type=5, link_count=i
        }
    }
    descriptor_length   = b110(buf,j); j+=2;
    k = j+descriptor_length;
    while (j<k) {
        j += do_desc(&buf[j],-1,-1);
    }

    return 0;
}
//-----------------------------------------------------------------------------
type_eventtable *find_crid_in_epg(int crid_type, char *crid, int *pevent_index)
{
    type_eventtable *et;
    int event_index;
    int j,k;

    //goto the start of the chain
    event_index = (pevent_index == 0) ? 0 : *pevent_index;
    if (event_index == 0) {
        k = 0;
        while (k < fw_max_events) {
            et = (type_eventtable *)((dword)fw_event_table + (event_index * fw_event_table_size));

            if ((et->char00 & 0xc0) == 0) {
                //find a valid entry
                event_index++;
            }
            else {
                //is this the start
                if (et->previous == 0xffff) {
                    break;
                }
                event_index = et->previous;
            }
            k++;
        }
        if (k == fw_max_events) {
            //epg empty
//            printf("\nepg empty");
            if (pevent_index) *pevent_index = 0;
            return 0;
        }
    }

    j = event_index;
    k = 0;
    //stop after N events
    while(k < fw_max_events) {
        et = (type_eventtable *)((dword)fw_event_table + (j * fw_event_table_size));

        //match episode crid
        if ((crid_type == 31)||(crid_type == 0)) {
            if (et->crid31_strlen) {
                if (memcmp(et->crid31_str, crid, et->crid31_strlen) == 0) {
                    if (pevent_index) *pevent_index = j;
                    return et;
                }
            }
        }
        //match series crid
        if ((crid_type == 32)||(crid_type == 0)) {
            if (et->crid32_strlen) {
                if (memcmp(et->crid32_str, crid, et->crid32_strlen) == 0) {
                    if (pevent_index) *pevent_index = j;
                    return et;
                }
            }
        }
        k++;

        j = et->next;
        if (j == 0xffff) break;
    }

//    printf("\ncannot find crid in epg");
    if (pevent_index) *pevent_index = fw_max_events;
    return 0;
}

type_eventtable *find_valid_crid_in_epg(int crid_type, char *crid, int *pevent_index)
{
    int event_index;
    type_eventtable *et;
    int nTvSvc, nRadioSvc;
    TYPE_TapChInfo chInfo;
    int i;


    event_index = (pevent_index == 0) ? 0 : *pevent_index;
    do {
        et = find_crid_in_epg(crid_type, crid, &event_index);             //31=episode, 32=series
        if (et) {
            //check that the channel in the epg is valid (sometimes it isn't ...)
            TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
            for (i=0; i<nTvSvc; i++) {
                TAP_Channel_GetInfo(SVC_TYPE_Tv, i, &chInfo);
                if (et->service_id == chInfo.svcId) {
                    return et;
                }
            }
            for (i=0; i<nRadioSvc; i++) {
                TAP_Channel_GetInfo(SVC_TYPE_Radio, i, &chInfo);
                if (et->service_id == chInfo.svcId) {
                    return et;
                }
            }
            event_index = et->next;
        }
    } while(et);
    return et;
}

void dump_event(type_eventtable *et, int event_index, char *str)
{
    int i;
    char event_name[0x100];
    word year;
    byte month, day, weekDay;
    dword utc, local;
    word start_date;
    byte start_hour, start_min;
    word end_date;
    byte end_hour, end_min;
    int nTvSvc, nRadioSvc;
    TYPE_TapChInfo chInfo;
    char str1[0x100];


    if (et->char00 & 0xc0) {
        utc = (et->start_date << 16) + (et->start_hour<<8) + et->start_min;
        local = exTAP_UTCtoLocalTime(utc);
        start_date = (local>>16)&0xffff;
        start_hour = (local>>8)&0xff;
        start_min = (local>>0)&0xff;
        TAP_ExtractMjd(start_date, &year, &month, &day, &weekDay);

        sprintf(str1, " ");
        strcpy(str, str1);
        if (start_date) {
            sprintf(str1, "%02d/%02d/%04d", day, month, year);
            strcat(str, str1);
        }
        sprintf(str1, ", %02d:%02d", start_hour, start_min);
        strcat(str, str1);

        utc = (et->end_date << 16) + (et->end_hour<<8) + et->end_min;
        local = exTAP_UTCtoLocalTime(utc);
        end_date = (local>>16)&0xffff;
        end_hour = (local>>8)&0xff;
        end_min = (local>>0)&0xff;
        TAP_ExtractMjd(end_date, &year, &month, &day, &weekDay);

        sprintf(str1, "-%02d:%02d", end_hour, end_min);
        strcat(str, str1);

        TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
        for (i=0; i<nTvSvc; i++) {
            TAP_Channel_GetInfo(SVC_TYPE_Tv, i, &chInfo);
            if (et->service_id == chInfo.svcId) {
                goto found;
            }
        }
        for (i=0; i<nRadioSvc; i++) {
            TAP_Channel_GetInfo(SVC_TYPE_Radio, i, &chInfo);
            if (et->service_id == chInfo.svcId) {
                goto found;
            }
        }
        goto not_found;

found:
//        sprintf(str1, " Ch. %d", chInfo.logicalChNum);
        sprintf(str1, " %s -", chInfo.chName);
        strcat(str, str1);

not_found:
        memcpy(event_name, et->event_name, et->event_name_length);
        event_name[et->event_name_length] = '\0';
        sprintf(str1, " %s", event_name);
        strcat(str, str1);
    }
}

//-----------------------------------------------------------------------------
int set_timer_by_crid(int crid_type, char *crid)
{
    type_eventtable *et;
    int event_index;
    int ret;

    event_index = 0;
    et = find_valid_crid_in_epg(crid_type, crid, &event_index);             //31=episode, 32=series
    if (!et) {
        //not in epg
        return -1;
    }

    //fill in firmware current event structure (0x81C040D8) with event from crid
    highlight_event(et);

    //record event
    ret = exTAP_CallFirmware((void *)fw_record_highlighted_event, 0,0,0,0,0,0,0,0,0,0,0);
    if ((ret != 0) &&(ret != 2)) {
        if (debug & 1) sprintf(strLog, "ERROR: fw_record_highlighted_event(), ret=%x", ret);    //2=menu, quit, 0=ok
        if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    }

    return ret;
}

//-----------------------------------------------------------------------------
//emulate fill 0x81C040D8 with current event
int highlight_event(type_eventtable *et)
{
    memcpy(fw_epg_highlighted_event, et, fw_event_table_size); //0x68

//firmware sets -0x7920($gp) ... not sure whether this is important; seems not
//    if (svcType) {
//        ret = getTVInfo(a0)
//        x = (word)ret[2];
//        >>6
//        *(word *)-0x7920($gp) = x;
//    }
//    else {
//        ret = getRadioInfo(a0);
//        x = (word)ret[2];
//        >>6
//        *(word *)-0x7920($gp) = x;
//    }

    *(dword *)(&fw_epg_highlighted_event->start_date) = exTAP_UTCtoLocalTime(*(dword *)(&fw_epg_highlighted_event->start_date));
    *(dword *)(&fw_epg_highlighted_event->end_date)   = exTAP_UTCtoLocalTime(*(dword *)(&fw_epg_highlighted_event->end_date));
    return 0;
}
//-----------------------------------------------------------------------------

void five_second_update(void)
{
    int ret;
    int main_channel;
    int svcType, svcNum;
    static int saved_svcType = -1;
    static int saved_svcNum = -1;
    TYPE_TapChInfo chInfo;
    static int busy = 0;

    //make sure we don't reenter
    if (busy) {
        if (debug & 1) sprintf(strLog, "ERROR: five_second_update(), 5 seconds rentered");
        if (debug & 1) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        return;
    }
    busy = 1;

    //check the hook is on
    if (demux_hook_enabled == 0) {
        pmt_handle = 0;
        rct_handle = 0;
        ret = apply_demux_hook();
        if (ret) {
            busy = 0;
            five_second_enabled = 0;
            return;
        }
        if (debug & 8) sprintf(strLog, "Demux Hook Enabled");
        if (debug & 8) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    }

    //operate on the main channel
    main_channel = ((*(volatile dword *)EMMA_TTSD_1ST_VIDEO_PID) >> 29)&3;

    //on channel change, forget the rct_pid
    TAP_Channel_GetCurrent(&svcType, &svcNum);
    if ((svcType != saved_svcType) || (svcNum  != saved_svcNum)) {
        related_content.rct_pid = 0;
        if (debug & 8) {
            sprintf(strLog,"five_second_update(), channel change detected, clearing rct_pid");
            log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        }
        if (debug & 0x10) {
            sprintf(strLog,"five_second_update(), channel change detected, clearing rct_pid");
            log_message(LOG_FILE, strLog, DELAY_SHORT);
        }

        if (rct_handle) {
            if (debug & 4) sprintf(strLog,"fw_free_filter(rct_handle = %x)", rct_handle);
            if (debug & 4) log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            exTAP_CallFirmware((void *)fw_free_filter, rct_handle, 0,0,0,0,0,0,0,0,0,0);
            rct_handle = 0;
        }
    }
    saved_svcType = svcType;
    saved_svcNum  = svcNum;

    //make sure we know the rct_pid
    if (related_content.rct_pid == 0) {
        ret = TAP_Channel_GetInfo(svcType, svcNum, &chInfo);
        if (debug & 8) sprintf(strLog, "pmt_pid = %04x", chInfo.pmtPid);
        if (debug & 8) log_message(LOG_SERIAL, strLog, DELAY_SHORT);

        //get+analyse the pmt
        ret = set_section_filter(main_channel, chInfo.pmtPid, 0x02, &pmt_handle);
    }
    //0xffff indicates rct not supported on this pmt
    else if (related_content.rct_pid != 0xffff){
        if (rct_handle == 0) {
            //get+analyse the rct
            ret = set_section_filter(main_channel, related_content.rct_pid, 0x76, &rct_handle);
        }
    }
    busy = 0;
}

//-----------------------------------------------------------------------------
void green()
{
    int ret;
    int i;
    type_eventtable *et;
    char str[0x80];

    if (debug & 0x10) {
        sprintf(strLog,"Green Button pressed");
        log_message(LOG_FILE, strLog, DELAY_SHORT);
    }
    //
    //Set a Timer recording from the crid
    //
    for (i=0; i<related_content.num_links; i++) {
       //try the series crid first
       if (related_content.link[i].term_id == 5) {
           ret = set_timer_by_crid(32, related_content.link[i].crid);
           if (ret == -1) {
                //series crid not matched; hopefully there will be an episode crid to try
                break;
            }
            if (ret == 0) {
                //timer set
                //if both a series, and episode crid exists, a Recording Option Dialog will popup
                //if only a series crid (can this happen?), the timer will be silently set, so a confirmation dialog would be nice
                return;
            }
            if (ret == 2) {
                //user quit Recording Option Dialog
                return;
            }
        }
    }

    for (i=0; i<related_content.num_links; i++) {
        //try the episode crid next
        if (related_content.link[i].term_id == 2) {
            ret = set_timer_by_crid(31, related_content.link[i].crid);
            if (ret == -1) {
                //crid not matched; timer not set, so a failure dialog would be nice
                sprintf(strLog, "Crid for %s not found in EPG. Recording not set", related_content.link[i].event_name);
                log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_MEDIUM);
                return;
            }
            if (ret == 0) {
                //timer set
                //if a series link exists, a Recording Option Dialog will popup
                //if a series link does not exist, the timer will be silently set, so a confirmation dialog would be nice
                et = find_valid_crid_in_epg(31, related_content.link[i].crid, NULL);
                dump_event(et, 0, str);
                sprintf(strLog, "Timer set: %s", str);
                log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_MEDIUM);
                return;
            }
            if (ret == 2) {
                //user quit Recording Option Dialog
                return;
            }
        }
    }

    //series crid not found - maybe series crid in rct, not (yet) in epg
    sprintf(strLog, "Crid for %s not found in EPG. Recording not set\nCrid=(%x)%s", related_content.link[0].event_name, related_content.link[0].term_id, related_content.link[0].crid);
    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_MEDIUM);
}

//-----------------------------------------------------------------------------
